import time
from collections import OrderedDict
from django.apps import apps
from django.db import models
from django.contrib.contenttypes.models import ContentType
from .utils.model import sorted_models, turn_off_auto_add, is_model_exists, get_generic_fk_info, get_generic_fk_fields
from .load import SeedLoader
from .fake import Guesser
from .fake.primary import get_int, get_str


class DekDjSeedException(Exception):
    pass


def create_project_entities(seed=None):
    seed = seed or time.time_ns()
    print(f'Current seed: {seed}')
    guesser = Guesser()
    for app_config in apps.get_app_configs():
        create_app_entities(app_config, guesser)


def create_app_entities(app_config, guesser):
    loader = SeedLoader.from_module(f'{app_config.name}.seed', guesser)
    for model in sorted_models(app_config):
        create_model_entities(model, loader)
    done = loader.seed_funcs.get('done')
    if done:
        done(guesser)


def create_model_entities(model_cls, loader):
    manager = model_cls.objects.db_manager()

    if not loader.get_model_config_value(model_cls, 'auto_add', False, model_cls):
        turn_off_auto_add(manager.model)

    generic_fk_info = get_generic_fk_info(model_cls)
    generic_fk_fields = get_generic_fk_fields(model_cls)

    for entity_index in range(loader.get_model_config_value(model_cls, 'number', 1, model_cls)):
        def get_custom_value(fd, dv, *args):
            data = {**faker_data, **faker_data_many_to_many, **custom_data}
            v = loader.get_model_field_value(model_cls, fd.name, data, dv, model_cls, entity_index, *args)
            cancel_map[fd.name] = data.pop('__cancel__', False)
            custom_data.update({k: v for k, v in custom_data.items() if k.startswith('__')})
            return dv if v is loader.EMPTY_VALUE else v

        cancel_map = {}
        faker_data = {}
        faker_data_many_to_many = OrderedDict()
        custom_data = {}

        count_check_unique = 0
        while True:
            faker_data = {}

            for gfk, ct, fk in generic_fk_info:
                field = model_cls._meta.get_field(gfk)
                value = get_custom_value(field, None)
                if value is not None:
                    faker_data[field.name] = value
                    # for is_model_exists
                    faker_data[ct] = ContentType.objects.get_for_model(value.__class__)
                    faker_data[fk] = value.pk

            for field in model_cls._meta.fields:
                if field.name in generic_fk_fields:
                    continue
                if field.primary_key:
                    if field.auto_created:
                        continue
                    if field.max_length is not None:
                        faker_data[field.name] = get_str()[: field.max_length]
                    else:
                        faker_data[field.name] = get_int()
                if field.has_default():
                    faker_data[field.name] = get_custom_value(field, field.get_default())
                    continue
                if isinstance(field, (models.OneToOneField, models.ForeignKey)):
                    array = get_custom_value(field, list(field.related_model.objects.all()))
                    if array:
                        faker_data[field.name] = array if isinstance(array, models.Model) \
                            else loader.guesser.random_sample(array, 1)[0]
                    continue
                formatter = loader.guesser.field_guess_format(field)
                if formatter:
                    faker_data[field.name] = get_custom_value(field, formatter())
                    continue
                # max length restriction check
                if field.max_length is not None:
                    faker_data[field.name] = faker_data[field.name][: field.max_length]

            if not is_model_exists(model_cls, faker_data):
                break
            count_check_unique += 1
            if count_check_unique >= 10000:
                raise DekDjSeedException(f'{model_cls} check unique do not passed, data: {faker_data}')

        for fk in generic_fk_fields:
            faker_data.pop(fk, None)

        if not any(cancel_map.values()):
            obj = manager.create(**faker_data)

            for field in model_cls._meta.many_to_many:
                if field.remote_field.through._meta.auto_created:
                    faker_data_many_to_many[field.name] = get_custom_value(
                        field,
                        list(field.related_model.objects.all()),
                        obj
                    )

            for field_name, related_obj_list in faker_data_many_to_many.items():
                for related_obj in related_obj_list:
                    getattr(obj, field_name).add(related_obj)



import re
import importlib
import inspect
from pathlib import Path
from dektools.yaml import yaml
from dektools.dict import assign
from django.conf import settings as dj_settings


class SeedException(Exception):
    pass


class EmptyValue:
    pass


class SeedLoader:
    EMPTY_VALUE = EmptyValue()
    seed_nest = 'seed'
    seed_prefix = seed_nest + '_'
    settings_filename = 'settings.yaml'
    default_seed = 'dekdjseed.core.seed'
    default_settings = dict(
        config=dict(number=10),
        models=[]
    )

    @classmethod
    def from_module(cls, module, guesser):
        def take_seeds(m):
            result = {}
            for k, v in vars(m).items():
                if k.startswith(cls.seed_prefix) and callable(v):
                    result[k[len(cls.seed_prefix):]] = v
            return result

        try:
            module_seed = importlib.import_module(module)
            filepath = Path(module_seed.__file__).parent / cls.settings_filename
        except ModuleNotFoundError:
            module_seed = None
            filepath = None
        try:
            module_seed_nest = importlib.import_module(cls.seed_nest + '.' + module)
            filepath_nest = (
                    Path(dj_settings.BASE_DIR) / cls.seed_nest / module.replace('.', '/') /
                    cls.settings_filename
            )
        except ModuleNotFoundError:
            module_seed_nest = None
            filepath_nest = None
        return cls(
            guesser,
            assign(
                yaml.load(Path(__file__).parent / cls.settings_filename),
                yaml.load(filepath) if filepath and filepath.exists() else {},
                yaml.load(filepath_nest) if filepath_nest and filepath_nest.exists() else {},
            ),
            assign(
                take_seeds(importlib.import_module(cls.default_seed)),
                take_seeds(module_seed) if module_seed else {},
                take_seeds(module_seed_nest) if module_seed_nest else {}
            )
        )

    def __init__(self, guesser, settings, seed_funcs=None):
        self.guesser = guesser
        self.seed_funcs = seed_funcs or {}
        self.models = {}
        self._parse(settings)

    def _parse(self, settings):
        def format_args(args, total):
            if len(args) == total:
                return args
            if len(args) < total:
                return args + [None] * (total - len(args))
            return args[:total]

        def transform_value(value):
            if isinstance(value, str):
                match = re.match(r'\$([a-zA-z_0-9]+)([(]*[\S\s]*[)]*)', value)
                if match:
                    func_name, args = match.groups()
                    func = self.seed_funcs.get(func_name)
                    if not func:
                        raise SeedException(f'Can not find func [{func_name}]')
                    args_total = len(inspect.signature(func).parameters)
                    try:
                        args = list(eval(args)) if args else []
                    except SyntaxError:
                        raise SeedException(f'SyntaxError [{value}] when eval [{args}]')
                    return lambda *x: func(*format_args([args, self.guesser, *x], args_total))
            return value

        for name, model_config in settings['models'].items():
            model_config = model_config or {}
            self.models[name] = dict(
                config={key: transform_value(value) for key, value in model_config.get('config', {}).items()},
                fields={key: transform_value(value) for key, value in model_config.get('fields', {}).items()}
            )

    def get_model(self, model_cls):
        result = dict()
        for cls in reversed(model_cls.mro()):
            result = assign(result, self.models.get(cls.__name__, {}))
        return result

    def get_model_config_value(self, model_cls, config_key, *args):
        return self.get_model_value(model_cls, ['config', config_key], *args)

    def get_model_field_value(self, model_cls, field_name, *args):
        return self.get_model_value(model_cls, ['fields', field_name], *args)

    def get_model_value(self, model_cls, keys, *args):
        item = self.get_model(model_cls)
        if item:
            cursor = item
            for key in keys[:-1]:
                cursor = cursor.get(key, {})
            value = cursor.get(keys[-1], self.EMPTY_VALUE)
            if value is not self.EMPTY_VALUE:
                if callable(value):
                    return value(*args)
                return value
        return self.EMPTY_VALUE



import string
from urllib.parse import urlparse


def seed_none(*args):
    return None


def seed_many_empty(args, guesser, data, value, model, index, instance):
    return []


def seed_default_number(args, guesser, value, model):
    return value


def seed_default_auto_add(args, guesser, value, model):
    return value


def seed_variable_name(args, guesser, data, value, model, index):
    args = args or (10, 20)
    return ''.join(guesser.random_sample(string.ascii_lowercase, guesser.random.randint(*args)))


def seed_class_name(args, guesser, data, value, model, index):
    args = args or (10, 20)
    return ''.join(
        guesser.random_sample(string.ascii_uppercase) +
        guesser.random_sample(string.ascii_lowercase, guesser.random.randint(*args))
    )


def seed_url(args, guesser):
    return guesser.fake.uri()


def seed_domain(args, guesser):
    return urlparse(guesser.fake.uri()).netloc



import string

string_queue = string.digits + string.ascii_letters
string_length = len(string_queue)

cursor_int = 0
cursor_str = [0]


def get_int():
    global cursor_int
    cursor_int += 1
    return cursor_int


def get_str():
    global cursor_str

    cursor = 0
    while cursor <= len(cursor_str) - 1:
        cursor_str[cursor] += 1
        if cursor_str[cursor] >= string_length:
            cursor_str[cursor] = 0
            cursor += 1
        else:
            break
    if cursor > len(cursor_str) - 1:
        cursor_str.append(0)

    return ''.join([string_queue[x] for x in cursor_str])



from toposort import toposort_flatten
from django.db.models import Q
from django.contrib.contenttypes.fields import GenericForeignKey


def turn_off_auto_add(model_cls):
    for field in model_cls._meta.fields:
        if getattr(field, "auto_now", False):
            field.auto_now = False
        if getattr(field, "auto_now_add", False):
            field.auto_now_add = False


def get_model_dependencies(models):
    dep_dict = {}
    dep_class_map = {}

    for model in models:
        dependencies = set()
        model_replacement = '{}.{}'.format(
            model.__module__,
            model.__name__
        )

        if model_replacement not in dep_class_map:
            dep_class_map[model_replacement] = model

        for field in model._meta.get_fields():
            if ((field.many_to_one is True or field.many_to_many is True or field.one_to_one is True) and
                    field.concrete and field.blank is False):

                related_model = field.related_model
                related_model_type = '{}.{}'.format(
                    related_model.__module__,
                    related_model.__name__
                )
                replacement = related_model_type

                if related_model_type not in dep_class_map:
                    dep_class_map[related_model_type] = related_model

                dependencies.add(replacement)

        dep_dict[model_replacement] = dependencies

    return dep_dict, dep_class_map


def sorted_models(app_config):
    dep_dict, dep_class_map = get_model_dependencies(app_config.get_models())
    return [dep_class_map[x] for x in toposort_flatten(dep_dict)]


def get_model_unique_fields(model_cls):
    result = []
    for field in model_cls._meta.fields:
        if not field.auto_created and field.unique:
            result.append([field])
    result.extend([[model_cls._meta.get_field(name) for name in pairs] for pairs in model_cls._meta.unique_together])
    return result


def is_model_exists(model_cls, data):
    def pkg_filter(k, v):
        if v is None:
            return f'{k}__isnull', True
        return k, v

    fields_list = get_model_unique_fields(model_cls)
    if not fields_list:
        return False
    query = Q()
    for fields in fields_list:
        kv = [pkg_filter(field.name, data.get(field.name)) for field in fields]
        query |= Q(**{item[0]: item[1] for item in kv})
    return model_cls.objects.filter(query).exists()


def get_generic_fk_info(model_cls):
    result = []
    for field in model_cls._meta.private_fields:
        if isinstance(field, GenericForeignKey):
            result.append((field.name, field.ct_field, field.fk_field))
    return result


def get_generic_fk_fields(model_cls):
    result = set()
    for item in get_generic_fk_info(model_cls):
        result.add(item[1])
        result.add(item[2])
    return result
